#!/usr/bin/python3

import pyvisa
import time
import climate_chamber_dummy


MONITOR_COMMAND_DELAY = 0.2  # program-related delay is 0.3
SETTING_COMMAND_DELAY = 0.5  # program-related delay is 1

'''
connect to either climate chamber or climate chamber dummy
depends on the hostname 
'''


def create_chamber(hostname, target_accuracy):
    if hostname == 'localhost':
        return climate_chamber_dummy.get_climate_chamber_dummy(target_accuracy)
    else:
        return ClimateChamber(hostname, target_accuracy)


class ClimateChamber:
    def __init__(self, ip_address, target_accuracy):
        """
                Implements the basic operation of the climate chamber

        """
        # open connection here

        # store the credentials so that methods can access them
        self.__ip_address = ip_address
        self.target_accuracy = target_accuracy

        # we try to connect to the climate chamber just to see if there is an error
        self.rm = pyvisa.ResourceManager()

        self.chamber = self.rm.open_resource("TCPIP0::" + str(self.__ip_address) + "::57732::SOCKET")
        self.chamber.write_termination = '\r\n'
        self.chamber.read_termination = '\r\n'
        self.chamber.timeout = 5000

    # the communication section

    def read_temperature(self):
        # data format: [current temp, set temp, upper limit, lower limit]
        # send the request to the chamber
        response = self.chamber.query("TEMP?", delay=MONITOR_COMMAND_DELAY)

        # self.my_socket.send(b'TEMP?\r\n')
        # temperature = self.my_socket.recv((1024).decode('utf-8'))

        # convert into float numbers
        temperature = [float(i) for i in response.split(sep=',')]
        return temperature

    def read_humidity(self):
        # data format: [current hum, set hum, upper limit, lower limit]
        # send the request to the chamber
        response = self.chamber.query("HUMI?", delay=MONITOR_COMMAND_DELAY)

        # convert into float numbers
        humidity = [float(i) for i in response.split(sep=',')]
        return humidity

    def set_temperature(self, temperature):
        # sets the temp of the chamber, (float) temperature
        response = self.chamber.query("TEMP, S" + str(temperature), delay=SETTING_COMMAND_DELAY)

        # response = (OK: TEMP, S30\r\n)
        return response

    def set_humidity(self, humidity):
        # sets the humidity of the chamber, (float) humidity
        response = self.chamber.query("HUMI, S" + str(humidity), delay=SETTING_COMMAND_DELAY)

        # response = (OK: HUMI, S30\r\n)
        return response

    def close(self):
        self.rm.close()

    # additional functionality

    def read_monitor(self):
        """ gets the chamber test area state
        output data format: [temp, humid, op-state, num. of alarms]
        """

        response = self.chamber.query("MON?", delay=MONITOR_COMMAND_DELAY)
        return response

    def set_temperature_limit(self, upper_limit, lower_limit):
        """
        Sets the upper and lower temperature limits for the chamber

        Parameters
        ----------
        upper_limit : int
            Temperature upper limit.
        lower_limit : int
            Temperature lower limit.

        Returns
        -------
        None.

        """
        response = self.chamber.query("TEMP, H" + str(upper_limit))
        response = self.chamber.query("TEMP, L" + str(lower_limit))

    def set_humidity_limit(self, upper_limit, lower_limit):
        """
        Sets the upper and lower humidity limits for the chamber

        Parameters
        ----------
        upper_limit : int
            humidity upper limit.
        lower_limit : int
            humidity lower limit.

        Returns
        -------
        None.

        """
        response = self.chamber.query("HUMI, H" + str(upper_limit))
        response = self.chamber.query("HUMI, L" + str(lower_limit))

    def read_mode(self):
        """ read the mode of the chamber:
        Panel power off "OFF"
        All operation stop "STANDBY"
        Constant operation "CONSTANT"
        Program/remote operation "RUN"
        """
        response = self.chamber.query("MODE?", delay=MONITOR_COMMAND_DELAY)
        return response

    def set_mode(self, mode):
        # sets the mode of the chamber:
        response = self.chamber.query("MODE, " + mode, delay=SETTING_COMMAND_DELAY)
        return response

    def _get_const(self, mode):
        response = self.chamber.query("CONSTANT SET?" + mode)
        return response

    def get_const(self):
        """
        combination of read_temperature() and read_humidity()
        """
        temp = self.read_temperature()
        hum = self.read_humidity()
        return [temp, hum]

    def set_const(self, setPoint):
        """

        Parameters
        ----------
        setPoint : tuple(float,float)
            A tuple of temperature and humidity as setpoint

        Returns
        -------
        None.

        """
        # setPoint = (temp,hum)
        self.set_temperature(setPoint[0])
        self.set_humidity(setPoint[1])

    def goto_const(self, setPoint, target_accuracy=(0.2, 1), soaking_time=60):
        """

        Parameters
        ----------
        setPoint :  tuple(float,float)
            A tuple of temperature and humidity as setpoint.
        target_accuracy : tuple(float,float), optional
            A tuple of temperature and humidity as accuracy. The default is (0.2, 1).
        soaking_time : int, optional
            Soaking time. The default is 60.

        Returns
        -------
        None.

        """

        current_milli_time = lambda: int(round(time.time() * 1000))

        # set temperature and humidity
        self.set_const(setPoint)

        # set mode to Const
        self.set_mode('CONSTANT')

        start_time = current_milli_time()
        # wait until setpoint is reached (+soaking time)
        while True:
            # get the actual temperature & humidity
            [temp, hum, mode, alarms] = self.read_monitor().split(',')

            # if it is not within the target range reset start time
            if not (setPoint[0] - target_accuracy[0] <= float(temp) <= setPoint[0] + target_accuracy[0]):
                start_time = current_milli_time()

            if not (setPoint[1] - target_accuracy[1] <= float(hum) <= setPoint[1] + target_accuracy[1]):
                start_time = current_milli_time()

            # if the value is stable long enough
            if (current_milli_time() - start_time) / 1000 > soaking_time:
                print('SOAKING FINISHED!')
                break

            print('Setpoint: ' + str(setPoint) + ' | Temp: ' + temp + ' Â°C'
                  + ' | Humid: ' + hum + '%'
                  + ' | time soaking:' + str((current_milli_time() - start_time) / 1000) + 's')

            time.sleep(1)

    # The climate chamber will instantly begin with the heat-up.
    # The Python process will instantly continue (so it will not wait until the target temperature is reached).
    # This function is blocking until the temperature and humidity setpoint have been reached

    def is_temperature_reached(self):

        # read_temperature returns measurement value and setpoint. Cache it, so we don't have to query the chamber twice
        temperature_response = self.read_temperature()
        if abs(temperature_response[0] - temperature_response[1]) < self.target_accuracy[0]:
            return True
        else:
            return False

    def is_humidity_reached(self):
        humidity_response = self.read_humidity()
        if abs(humidity_response[0] - humidity_response[1]) < self.target_accuracy[1]:
            return True
        else:
            return False

    def wait_for_setpoints_reached(self):
        while (not self.is_humidity_reached()) or (not self.is_temperature_reached()):
            time.sleep(10)
            break
    
    def get_heater_percentage(self):
        
        response = self.chamber.query("%?", delay=MONITOR_COMMAND_DELAY)
        
        heater_param = response.split(sep=',')
        heater_count = int(heater_param[0])
        perc_temp_heater = float(heater_param[1])
        perc_hum_heater = float(heater_param[2])
        
        return perc_temp_heater, perc_hum_heater
            
            
            

    def __del__(self):
        self.close()
